/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.activemq.apollo.cli.commands

import org.apache.felix.gogo.commands.{Action, Option => option, Argument => argument, Command => command}
import org.apache.activemq.apollo.broker.{Broker, ConfigStore}
import org.apache.activemq.apollo.util.FileSupport._
import org.apache.activemq.apollo.cli.Apollo
import org.apache.felix.service.command.CommandSession
import org.fusesource.hawtdispatch._
import org.apache.activemq.apollo.util.{FileMonitor, ServiceControl}
import org.apache.log4j.PropertyConfigurator
import java.io.{FileInputStream, File}
import java.util.logging.LogManager
import org.apache.activemq.apollo.dto.BrokerDTO
import collection.mutable.ListBuffer

/**
 * The apollo run command
 */
@command(scope="apollo", name = "run", description = "runs the broker instance")
class Run extends Action {

  def system_dir(name:String) = {
    val base_value = System.getProperty(name)
    if( base_value==null ) {
      sys.error("The the %s system property is not set.".format(name))
    }
    val file = new File(base_value)
    if( !file.isDirectory  ) {
      sys.error("The the %s system property is not set to valid directory path %s".format(name, base_value))
    }
    file
  }

  def execute(session: CommandSession):AnyRef = {

    try {

      val base = system_dir("apollo.base")
      val etc: File = base / "etc"

      val log4j_config = etc / "log4j.properties"
      PropertyConfigurator.configure(log4j_config.getCanonicalPath)

      val conf = etc / "apollo.xml"

      if( !conf.exists ) {
        sys.error("Configuration file'%s' does not exist.\n\nTry creating a broker instance using the 'apollo create' command.".format(conf));
      }

      if( System.getProperty("java.security.auth.login.config")==null ) {
        val login_config = etc / "login.config"
        if( login_config.exists ) {
          System.setProperty("java.security.auth.login.config", login_config.getCanonicalPath)
        }
      }

      val tmp = base / "tmp"
      tmp.mkdirs

      Apollo.print_banner(session.getConsole)

      def println(value:String) = session.getConsole.println(value)

      // Load the configs and start the brokers up.
      println("Loading configuration file '%s'.".format(conf))

      val broker = new Broker()

      val validation_messages = ListBuffer[String]()
      try {
        broker.config = ConfigStore.load(conf, validation_messages += _)
      } finally {
        if( !validation_messages.isEmpty) {
          println("")
          println("Broker configuration file failed the following validations:")
          validation_messages.foreach{ v =>
            println("")
            println("  "+v)
          }
          println("")
        }
      }

      if( broker.config.validation == "strict" && !validation_messages.isEmpty) {
        session.getConsole.println("Strict validation was configured, shutting down")
        return null
      }

      broker.tmp = tmp
      broker.start()

      val broker_config_monitor = new FileMonitor(conf,broker.dispatch_queue {
        broker.console_log.info("Reloading configuration file '%s'.".format(conf))
        broker.update(ConfigStore.load(conf, x=> broker.console_log.info(x) ), ^{
        })
      })
      val log4j_config_monitor = new FileMonitor(log4j_config, {
        PropertyConfigurator.configure(log4j_config.getCanonicalPath)
      })

      var jul_config = etc / "jul.properties"
      val jul_config_monitor = if ( jul_config.exists()) {
        new FileMonitor(jul_config, {
          using(new FileInputStream(jul_config)) { is =>
            LogManager.getLogManager.readConfiguration(is)
          }
        })
      } else {
        null
      }

      if(jul_config_monitor!=null) jul_config_monitor.start
      log4j_config_monitor.start
      broker_config_monitor.start

      Runtime.getRuntime.addShutdownHook(new Thread(){
        override def run: Unit = {
          if(jul_config_monitor!=null) ServiceControl.stop(jul_config_monitor, "stopping config monitor")
          ServiceControl.stop(log4j_config_monitor, "stopping config monitor")
          ServiceControl.stop(broker_config_monitor, "stopping config monitor")
          ServiceControl.stop(broker, "stopping broker")
        }
      })

      // wait forever...  broker will system exit.
      this.synchronized {
        this.wait
      }

    } catch {
      case x:Helper.Failure=>
        sys.error(x.getMessage)
    }
    null
  }

}
